/*
gfu - gophermap format utility

`gfu` manipulates gophermaps (gopher menus). It is intended to be used as part
of an automation chain for managing a gopherhole using maps as the main
doctype, easily allowing any document to contain links.

`gfu` can:
 - Convert all lines in a file that are not valid gopher links into gopher info
 text (item type 'i') lines
 - Deconstructing all gopher info text lines in a file back to plain text for
 easy editing
 - Insert the contents of a header file to the start of a document
 - Append the contents of a footer file to the end of a document

Documentation

For information on using `gfu`, see the help information: gfu --help

Information on building, downloading and installing `gfu` can be found at:
https://tildegit.org/sloum/gfu

Further information on gfu can be found at the `gfu` homepage:
https://rawtext.club/~sloum/gfu.html
*/
package main

import (
	"bufio"
	"flag"
	"fmt"
	"io"
	"log"
	"os"
	"regexp"
	"strings"
)

//version information, set by ldflags with some default fallbacks
var (
	version = "not set"
	build   = "not set"
)

//command line flag global variables
var (
	deconstructInfoTextLines bool
	header                   string
	footer                   string
	inplace                  bool
	outfile                  string
	printVersion             bool
)

//regex global variable that identifies the format of gopher menu lines
var reGopherLines = regexp.MustCompile(`.+\t.*\t.*\t.*`)

func buildInfoText(ln string) string {
	var out strings.Builder
	out.WriteString("i")
	out.WriteString(ln)
	out.WriteString("\tfalse\tnull.host\t1")
	return out.String()
}

func deconstructInfoText(ln string) string {
	text := strings.SplitN(ln, "\t", 2)
	infotext := text[0]
	if len(infotext) > 1 {
		return infotext[1:]
	}
	return ""
}

//takes a string representing the file path, opens the file, then passes the
//file io for reading/processing. returns data as a string and any error
//information.
func readFile(path string) (s string, err error) {
	file, err := os.Open(path)
	if err != nil {
		return
	}

	defer file.Close()

	s, err = processFile(file)
	if err != nil {
		return
	}

	return s, err
}

//takes file io, reads and processes each line according to the
//deconstructInfoTextLines flag. returns processed data as a string and any
//error information.
func processFile(file io.Reader) (s string, err error) {
	scanner := bufio.NewScanner(file)
	var sb = strings.Builder{}

	if !deconstructInfoTextLines {
		for scanner.Scan() {
			l := scanner.Text()
			if exp := reGopherLines.MatchString(l); exp {
				sb.WriteString(l)
			} else {
				sb.WriteString(buildInfoText(l))
			}
			sb.WriteRune('\n')
		}
	} else {
		for scanner.Scan() {
			l := scanner.Text()
			if exp := reGopherLines.MatchString(l); exp && l[0] == 'i' {
				sb.WriteString(deconstructInfoText(l))
			} else {
				sb.WriteString(l)
			}
			sb.WriteRune('\n')
		}
	}
	if err = scanner.Err(); err != nil {
		return
	}

	return sb.String(), nil
}

//takes a file path as a string and some data as a string. writes the data to
//the specified file. returns any error information.
func writeFile(path string, outFile string) (err error) {
	file, err := os.OpenFile(path, os.O_WRONLY|os.O_CREATE|os.O_TRUNC, 0644)
	if err != nil {
		return
	}
	defer file.Close()
	file.WriteString(outFile)
	return nil
}

//configure command line flags
func init() {
	flag.BoolVar(&deconstructInfoTextLines, "d", false, "")
	flag.BoolVar(&deconstructInfoTextLines, "deconstruct", false, "Deconstruct a gophermap's info text lines back to plain text")
	flag.StringVar(&header, "h", "", "")
	flag.StringVar(&header, "header", "", "Specify a file to be inserted as header content.")
	flag.StringVar(&footer, "f", "", "")
	flag.StringVar(&footer, "footer", "", "Specify a file to be appended as footer content.")
	flag.BoolVar(&inplace, "i", false, "")
	flag.BoolVar(&inplace, "in-place", false, "Processing is done in-place, saving any changes to the specified input file.")
	flag.StringVar(&outfile, "o", "", "")
	flag.StringVar(&outfile, "outfile", "", "Save processed output to the specified file.")

	flag.BoolVar(&printVersion, "v", false, "")
	flag.BoolVar(&printVersion, "version", false, "Output version information and quit")
}

//PrintHelp produces a nice display message when the --help flag is used
func PrintHelp() {
	art := `gfu - gophermap formatting utility

syntax: gfu [flags...] [filepath]

example: gfu -d ~/gopher/phlog/gophermap

default
	Convert plain text lines to gophermap info text (item type 'i') lines
`
	fmt.Fprint(os.Stdout, art)
	flag.PrintDefaults()
}

func main() {
	// read and process command line options/flags, returning requested
	// information or handling bad combinations
	flag.Parse()
	flag.Usage = PrintHelp

	if printVersion {
		fmt.Printf("gfu - gophermap format utility - version %s build %s\n", version, build)
		os.Exit(0)
	}
	var infile string
	if l := len(flag.Args()); inplace && l != 1 {
		fmt.Fprintf(os.Stderr, "Incorrect number of arguments. Expected 1, got %d\n", l)
		os.Exit(1)
	} else if l == 1 {
		infile = flag.Args()[0]
	}
	if inplace && outfile != "" {
		fmt.Fprintf(os.Stderr, "--in-place and --output options are mutually exclusive\n")
		os.Exit(1)
	}

	// read each file or stream, in the order they will be arranged for output
	// append processed output
	var output = strings.Builder{}
	if header != "" {
		s, err := readFile(header)
		if err != nil {
			log.Fatalln("Error while reading file -", err)
		}
		output.WriteString(s)
	}
	if infile != "" {
		s, err := readFile(infile)
		if err != nil {
			log.Fatalln("Error while reading file -", err)
		}
		output.WriteString(s)
	} else {
		s, err := processFile(os.Stdin)
		if err != nil {
			log.Fatalln("Error while reading stdin -", err)
		}
		output.WriteString(s)
	}
	if footer != "" {
		s, err := readFile(footer)
		if err != nil {
			log.Fatalln("Error while reading file -", err)
		}
		output.WriteString(s)
	}

	//output data to stdout or file
	switch {
	case inplace:
		err := writeFile(infile, output.String())
		if err != nil {
			log.Fatalln("Error while writing file -", err)
		}
	case outfile != "":
		err := writeFile(outfile, output.String())
		if err != nil {
			log.Fatalln("Error while writing file -", err)
		}
	default:
		fmt.Print(output.String())
	}

	//the end
	os.Exit(0)
}
